unit HiRDRemover;    // Removable Device Remover

interface

uses kol, Share, Debug, Windows;

const

 IOCTL_STORAGE_MEDIA_REMOVAL = 2967556;
 IOCTL_STORAGE_EJECT_MEDIA = 2967560;

 FSCTL_LOCK_VOLUME = 589848; 
 FSCTL_DISMOUNT_VOLUME = 589856;
  
type

 PDevBroadcastHdr = ^DEV_BROADCAST_HDR;
 DEV_BROADCAST_HDR = packed record
   dbch_size: DWORD;
   dbch_devicetype: DWORD;
   dbch_reserved: DWORD;
 end;
 
 PREVENT_MEDIA_REMOVAL = record
   PreventMediaRemoval: Boolean;
 end;   

 THiRDRemover = class(TDebug)
   public
    _prop_Drive: String;
    _data_Drive: THI_Event;
    _event_onRemove: THI_Event;
    _event_onError: THI_Event;
    procedure _work_doRemove(var _Data:TData; Index:word); 
 end;

implementation

function OpenVolume(DriveLetter: Char): THandle;
var
  szRoot:   PChar;
  szVolume: PChar;
  dwAccess: DWORD;
begin

  szRoot := PChar(DriveLetter+':\'); 

  case GetDriveType(szRoot) of
    DRIVE_REMOVABLE: dwAccess := GENERIC_READ or GENERIC_WRITE;
    DRIVE_CDROM:     dwAccess := GENERIC_READ;
    DRIVE_FIXED:     dwAccess := GENERIC_READ or GENERIC_WRITE;
    else begin Result := INVALID_HANDLE_VALUE; Exit; end;
  end;

  szVolume := PChar('\\.\'+DriveLetter+':'); 
  Result := CreateFile(szVolume, dwAccess, FILE_SHARE_READ or FILE_SHARE_WRITE,
                       nil, OPEN_EXISTING, 0, 0);
end;

function LockVolume(hVolume: THandle): LongBool;
var
  dwRet:   DWORD;
  dwSleep: DWORD;
  I:       Integer;
begin
  Result  := False;
  dwSleep := 100; //LOCK_TIMEOUT div LOCK_RETRY;

  for i := 0 to 10 {LOCK_RETRY } do
  begin
    if DeviceIoControl(hVolume, FSCTL_LOCK_VOLUME, nil, 0, nil, 0, dwRet, nil) then
    begin
      Result := True;
      Exit;
    end;

    Sleep(dwSleep);
  end;
end;        

function DismountVolume(hVolume: THandle): LongBool;
var
  dwRet: DWORD;
begin
  Result := DeviceIoControl(hVolume, FSCTL_DISMOUNT_VOLUME, nil, 0, 
                            nil, 0, dwRet, nil);
end;    

function AutoEjectMedia(hVolume: THandle): LongBool;
var
  dwRet: DWORD;
begin
  Result := DeviceIoControl(hVolume, IOCTL_STORAGE_EJECT_MEDIA, nil, 0, 
                            nil, 0, dwRet, nil);
end;

function PreventVolumeRemoval(hVolume: THandle; Prevent: Boolean): LongBool;
var
  dwRet: DWORD;
  PRM:   PREVENT_MEDIA_REMOVAL;
begin
  PRM.PreventMediaRemoval := Prevent;
  Result := DeviceIoControl(hVolume, IOCTL_STORAGE_MEDIA_REMOVAL, @PRM,
                            SizeOf(PREVENT_MEDIA_REMOVAL), nil, 0, dwRet, nil);
end;

procedure THiRDRemover._work_doRemove;
var
  DriveLetter: String;
  hVol: THandle;
begin
  DriveLetter:= ReadString(_Data, _data_Drive, _prop_Drive);
  hVol := OpenVolume(DriveLetter[1]);
  
  if hVol = INVALID_HANDLE_VALUE then 
    _Hi_OnEvent(_event_onError)
   else begin
    if (LockVolume(hVol) and DismountVolume(hVol) and 
                PreventVolumeRemoval(hVol, False)) then AutoEjectMedia(hVol);
    if CloseHandle(hVol) then _Hi_OnEvent(_event_onRemove, DriveLetter);
  end; 
end;

end.

